{***********************************************************************}
{ Unite:   ByFiles							}
{ Version: 1.00 beta 1							}
{									}
{ Developer:								}
{   African Dancer (dcwang@interchange.ubc.ca)				}
{									}
{ DevLog:								}
{   May 25	+ special folders (Dancer)				}
{   Jun 25	+ multiple special folders (Dancer)			}
{   Dec 07	+ finally got it to work (though still unstable)	}
{									}
{***********************************************************************}

unit Byfiles;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, StdCtrls, ExtCtrls, DsgnIntf;

type
  TFileFolderWin = class(TForm)
    BtnNew: TButton;
    BtnDel: TButton;
    BtnPost: TButton;
    BtnUndo: TButton;
    ListBox: TListBox;
    Image: TImage;
    ImageOpen: TOpenDialog;
    ImageHolder: TGroupBox;
    FilterEdit: TEdit;
    FilterLabel: TLabel;
    BtnOK: TButton;
    Label1: TLabel;
    procedure LBDraw(Ctrl: TWinControl; Idx: Integer; R: TRect; State: TOwnerDrawState);
    procedure ImageClick(Sender: TObject);
    procedure New(Sender: TObject);
    procedure Delete(Sender: TObject);
    procedure Post(Sender: TObject);
    procedure Undo(Sender: TObject);
    procedure LBKeyPress(Sender: TObject; var Key: Char);
    procedure DestroyForm(Sender: TObject);
  end;

  TByFileFolder = class(TPersistent)
  private
    FFilter : string;
    FIcon   : TBitmap;
    procedure SetIcon(value: TBitmap);
  protected
    procedure WriteData(Stream: TStream);
    procedure ReadData(Stream: TStream);
  public
    constructor Create; virtual;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    property Filter: string read FFilter write FFilter;
    property Icon: TBitmap read FIcon write SetIcon;
  end;

  TByFileFolders = class(TPersistent)
  private
    FItems: Array[0..9] of TByFileFolder;
    FCount: Integer;
    function GetItem(Index: Integer): TByFileFolder;
    procedure SetItem(Index: Integer; Value: TByFileFolder);
    function GetCount: integer;
  protected
    procedure WriteData(Stream: TStream);
    procedure ReadData(Stream: TStream);
  public
    procedure Assign(Source: TPersistent); override;
    procedure Add(value: TByFileFolder);
    procedure New(filter: string; icon: TBitmap);
    procedure Clear;

    constructor Create; virtual;
    destructor Destroy; override;

    property Count: integer read GetCount;
    property Items[Index: Integer]: TByFileFolder read GetItem write SetItem; default;
  end;

  TByFileFolderEdit = class(TClassProperty)
  public
    function GetAttributes: TPropertyAttributes; override;
    procedure Edit; override;
  end;

const
  { TByFileNode type }
  FONDiskDrv = $0010;
  FONUnknown = $0000 or FONDiskDrv;
  FONFloppy  = $0002 or FONDiskDrv;
  FONFixed   = $0004 or FONDiskDrv;
  FONCDROM   = $0005 or FONDiskDrv;
  FONNetDrv  = $0006 or FONDiskDrv;

  FONFolder  = $0020;
  FONNormal  = $0001 or FONFolder;
  FONSpecial = $0003 or FONFolder;

  { Error code for TByFileNode.Refresh }
  FENoError       = 0; { Self-explainatory }
  FEDirNotExist   = 1; { This directory no longer exists }
  FENoSubDirExist = 2; { FHasSubDir = true but directory has no sub directory }
  FESubDirExist   = 3; { FHasSubDir = false but the directory has sub directories }
  FEDriveNotReady = 4;

type
  TByFileNode = class(TObject)
    FDirName,
    FFullPath : string;
    FLevel : integer;
    FHasSubDir : boolean;
    FIsOpen : boolean;
    FType : integer;
    FFilterIndex: integer;
    procedure SetFilterIndex(value: integer);
    procedure SetOpen(value: boolean);
  public
    constructor Create(APath, AName: string; ALevel:integer; AType: integer);
    function Refresh: integer;

    property DirName  : string read FDirName;
    property FullPath : string read FFullPath;
    property FolderType: integer read FType;
    property HasSubDir: boolean read FHasSubDir;
    property IsOpen: boolean read FIsOpen write SetOpen;
    property Level: integer read FLevel;
    property FilterIndex : integer read FFilterIndex write setFilterIndex;
  end;

  TByFileDirectory = class(TCustomListBox)
  private
    FOpen, FClose, FFloppy, FFixed, FCDRom, FRemote : TBitmap;
    FDirChange: TNotifyEvent;

    FFileFolders: TByFileFolders;
    function GetDirectory: string;
    procedure SetDirectory(value: string);
    procedure SetFileFolders(value: TByFileFolders);
    procedure CNDrawItem(var msg: TWMDrawItem); message CN_DRAWITEM;

    procedure ReadData(Stream: TStream);
    procedure WriteData(Stream: TStream);
  protected
    procedure DefineProperties(Filer: TFiler); override;

    function Collapse: Boolean;
    function Expand: Boolean;
    procedure SetScrollBar;
    procedure DrawItem(Ctrl: TWinControl; Idx: Integer; R: TRect; State: TOwnerDrawState);
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure DblClick; override;
    procedure Click; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure CreateWnd; override;
    property Directory: string read GetDirectory write SetDirectory;
  published
    property Align;
    property Color;
    property Ctl3D;
    property BorderStyle;
    property ParentColor;
    property ParentCtl3D;
    property FileFolders: TByFileFolders read FFileFolders write SetFileFolders;
    property TabStop;
    property TabOrder;
    property OnDirectoryChange: TNotifyEvent read FDirChange write FDirChange;
    property OnClick;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnKeyDown;
    property OnKeyUp;
  end;

function ShortenFileDir(s: string; MaxLenght: integer) : string;
function FileExists(name: string): boolean;
function DriveReady(const DriveLetter: Char): boolean;

procedure Register;

implementation

uses ByConst;

{$R *.res}
{$R *.dfm}

procedure Register;
begin
  RegisterComponents('Beyond 98', [TByFileDirectory]);
  RegisterPropertyEditor(TypeInfo(TByFileFolders), nil, '', TByFileFolderEdit);
end;

function TByFileFolderEdit.GetAttributes: TPropertyAttributes;
begin
  Result := [paDialog];
end;

procedure TByFileFolderEdit.Edit;
var
  Editor: TFileFolderWin;
  FileFolders: TByFileFolders;
  last: integer;
  idx : integer;
  bmp : TBitmap;
begin
  inherited Edit;

  Editor := TFileFolderWin.Create(nil);
  try
    FileFolders := TByFileFolders(GetOrdValue);

    last := fileFolders.count - 1;
    if last >= 0 then
    begin
      for idx := 0 to last do
	with FileFolders.Items[idx] do
        begin
          bmp := TBitmap.Create;
          bmp.Assign(Icon);
	  Editor.ListBox.Items.AddObject(Filter, bmp);
        end;

      Editor.ListBox.ItemIndex := 0;
      Editor.ImageHolder.Enabled := true;
      Editor.FilterEdit.Enabled := true;
    end;

    if Editor.ShowModal = mrOk then
    begin
      FileFolders.Clear;
      last := editor.listBox.items.count - 1;

      if (last >= 0) then
	for idx := 0 to last do
	  with Editor.ListBox do
	    FileFolders.New(Items[idx], TBitmap(Items.Objects[idx]));
    end;

  finally
    Editor.free;
  end;
end;

{******************************************************************************}

procedure TFileFolderWin.LBDraw(Ctrl: TWinControl; Idx: Integer; R: TRect; State: TOwnerDrawState);
var
  bmp: TBitmap;
  w, h: integer;
  Transparent: TColor;
  Dest, Source : TRect;
begin
  ListBox.Canvas.FillRect(R);
  if ListBox.Items.Objects[Idx] <> nil then
  begin
    Bmp := TBitmap(ListBox.Items.Objects[Idx]);
    w := Bmp.Width - 1;
    h := Bmp.Height - 1;
    Transparent := bmp.canvas.pixels[w,h];
    Dest := Rect(R.left + 2, R.Top + 1, R.left + 18, R.Bottom -1);
    Source := Rect(0, 0, w, h);
    ListBox.Canvas.BrushCopy(Dest, Bmp, Source, transparent);
  end;
  ListBox.Canvas.TextOut(20, r.top + 1,ListBox.Items[idx]);
end;

procedure TFileFolderWin.ImageClick(Sender: TObject);
begin
  if ImageOpen.Execute then Image.Picture.Bitmap.LoadFromFile(ImageOpen.fileName);
end;

procedure TFileFolderWin.New(Sender: TObject);
var
  bmp: TBitmap;
begin
  bmp := TBitmap.Create;
  ListBox.Items.AddObject('*.*', bmp);
  if ListBox.Items.Count = 1 then
  begin
    ListBox.ItemIndex := 0;
    ImageHolder.enabled := true;
    FilterEdit.enabled := true;
  end;
end;

procedure TFileFolderWin.Delete(Sender: TObject);
var
  idx: integer;
begin
  idx := ListBox.ItemIndex;
  if idx > -1 then
  begin
    TBitmap(ListBox.Items.Objects[idx]).free;
    ListBox.Items.Delete(idx);

    if ListBox.Items.Count = 0 then
    begin
      ImageHolder.enabled := false;
      FilterEdit.enabled := false;
    end;
  end;
end;

procedure TFileFolderWin.Post(Sender: TObject);
var
  idx: integer;
begin
  idx := ListBox.ItemIndex;
  if idx > -1 then
  begin
    TBitmap(ListBox.Items.Objects[idx]).Assign(Image.Picture.Bitmap);
    ListBox.Items[idx] := FilterEdit.text;
  end;
end;

procedure TFileFolderWin.Undo(Sender: TObject);
var
  idx: integer;
begin
  idx := ListBox.ItemIndex;
  if idx > -1 then
  begin
    Image.Picture.Bitmap.Assign(TBitmap(ListBox.Items.Objects[idx]));
    FilterEdit.text := ListBox.Items[idx];
  end;
end;

procedure TFileFolderWin.LBKeyPress(Sender: TObject; var Key: Char);
begin
  Undo(nil);
end;

procedure TFileFolderWin.DestroyForm(Sender: TObject);
var
  i, last: integer;
begin
  last := ListBox.Items.Count - 1;
  if last >= 0 then
    for i := last downto 0 do
        TBitmap(ListBox.Items.Objects[i]).free;
end;

{*****************************************************************************}

constructor TByFileFolder.Create;
begin
  inherited Create;
  FFilter := '*.*';
  FIcon := TBitmap.Create;
end;

destructor TByFileFolder.Destroy;
begin
  FIcon.free;
  inherited Destroy;
end;

procedure TByFileFolder.Assign(Source: TPersistent);
var
  obj:TByFileFolder;
begin
  if Source is TByFileFolder then
  begin
    obj := TByFileFolder(Source);
    FIcon.Assign(obj.Icon);
    FFilter := obj.Filter;
    Exit;
  end;
  inherited Assign(Source);
end;

procedure TByFileFolder.SetIcon(value: TBitmap);
begin
  FIcon.Assign(value);
{  if not FIcon.Empty then FIcon.Dormant; }
end;

type
  TByFileFolderInfo = packed record
    Filter: String[40];
    BmpSize: LongInt;
  end;

procedure TByFileFolder.WriteData(Stream: TStream);
var
  Data: TByFileFolderInfo;
  BmpStream: TMemoryStream;
begin
  BmpStream := TMemoryStream.Create;
  try
    Data.Filter := Copy(FFilter, 1, 40);

    if FIcon.Empty
    then
      Data.BmpSize := 0
    else
      begin
	FIcon.SaveToStream(BmpStream);
	Data.BmpSize := BmpStream.Size;
	BmpStream.Position := 0;
      end;

    Stream.WriteBuffer(Data, SizeOf(Data));
    if Data.BmpSize > 0 then
      Stream.CopyFrom(BmpStream, Data.BmpSize);
  finally
    BmpStream.Free;
  end;
end;

procedure TByFileFolder.ReadData(Stream: TStream);
var
  Data: TByFileFolderInfo;
  BmpStream: TMemoryStream;
begin
  Stream.ReadBuffer(Data, SizeOf(Data));
  FFilter := Data.Filter;
  if Data.BmpSize > 0
  then
    begin
      BmpStream := TMemoryStream.Create;
      try
	BmpStream.CopyFrom(Stream, Data.BmpSize);
	BmpStream.Position := 0;
	FIcon.LoadFromStream(BmpStream);
      finally
       BmpStream.Free;
      end;
    end
  else
    FIcon.Assign(nil);
end;

{*****************************************************************************}

constructor TByFileFolders.Create;
begin
  inherited Create;
  FCount := 0;
end;

destructor TByFileFolders.Destroy;
begin
  clear;
  inherited Destroy;
end;

procedure TByFileFolders.WriteData(Stream: TStream);
var idx: Integer;
begin
  Stream.WriteBuffer(FCount, SizeOf(FCount));
  for Idx := count - 1 downto 0 do
    FItems[idx].WriteData(Stream);
end;

procedure TByFileFolders.ReadData(Stream: TStream);
var
   count, idx: Integer;
   FileFolder: TByFileFolder;
begin
  clear;
  Stream.ReadBuffer(count, SizeOf(count));
  for idx := count - 1 downto 0 do
  begin
    FileFolder := TByFileFolder.Create;
    FileFolder.ReadData(Stream);
    FItems[idx] := FileFolder;
  end;
  fcount := count;
end;

function TByFileFolders.GetItem(Index: Integer): TByFileFolder;
begin
  GetItem := FItems[index];
end;

procedure TByFileFolders.SetItem(Index: Integer; Value: TByFileFolder);
begin
{  if (Index > -1) and (Index < FItems.Count) then }
  FItems[Index].Assign(Value)
end;

procedure TByFileFolders.Assign(Source: TPersistent);
var
  index, count: integer;
  FileFolders : TByFileFolders;
  FileFolder : TByFileFolder;
begin
  if Source is TByFileFolders then
  begin
    FileFolders := TByFileFolders(Source);
    clear;
    count := FileFolders.Count - 1;
    fcount := count;
    if count > 0 then
      for index := count - 1 downto 0 do
      begin
        FileFolder := TByFileFolder.Create;
        FileFolder.Assign(FileFolders.items[index]);
	FItems[index] := FileFolder;
      end;
  end;
end;

function TByFileFolders.GetCount: integer;
begin
  GetCount := fcount;
end;

procedure TByFileFolders.Add(value: TByFileFolder);
begin
  if Value <> nil then
  begin
    FItems[FCount] := Value;
    Inc(FCount);
  end;
end;

procedure TByFileFolders.New(filter:string; icon: TBitmap);
var obj: TByFileFolder;
begin
  obj := TByFileFolder.Create;
  obj.Icon := icon;
  obj.filter := filter;
  FItems[FCount] := obj;
  Inc(FCount);
end;

procedure TByFileFolders.Clear;
var index : integer;
begin
  if FCount > 0 then
  begin
    for index := FCount - 1 downto 0 do
       FItems[index].Free;
    FCount := 0;
  end;
end;

{*****************************************************************************}

{ delphiFAQArchive/1595.html.}
function DriveReady(const DriveLetter: Char): Boolean;
var
  OldErrorMode: Word;
  OldDirectory: string;
begin
  OldErrorMode:= SetErrorMode(SEM_NOOPENFILEERRORBOX);
  GetDir(0, OldDirectory);
  {$I-}
  ChDir(DriveLetter + ':\');
  {$I+}
  if IoResult <> 0 then
    Result:= False
  else
    Result:= True;
  ChDir(OldDirectory);
  SetErrorMode(OldErrorMode);
end;

{ this function is not ready }
function ShortenFileDir(s: string; MaxLenght: integer) : string;
var
  Dir : string;
  FName : string[12];
  L : byte;
begin
  L := Length(s);

  if L > MaxLenght then
    begin
      FName := ExtractFileName(s);
      Dir := copy(s, 0, L - Length(FName));
      L := length(Dir);
{      if length(dir) <> i then dir:=dir+'\';
        shortenfilename := copy(dir, i+1 , length(dir) -i-1)+name+ext;}
    end;

  Result := s;
end;

{ use this so that I dont have to use Files unit and I will not have
  TFileDirectory and TFileListBox in my application }
function FileExists(name: string): boolean;
const
  mask: integer = faArchive + faDirectory + faReadOnly;
var
  f: TSearchRec;
  err: integer;
begin
  Result := false;
  if (name[2] = ':') and (length(name) = 2) then name := name + '\*.*';

  err := FindFirst(name, mask, f);
  while err = 0 do
    if {((f.Attr and mask) = mask) and }(f.Name[1] <> '.') then
      begin
        err := 1;
        Result := true;
      end else
      err := findnext(f);
end;

{*****************************************************************************}

constructor TByFileNode.Create(APath, AName: string; ALevel:integer; AType: integer);
var
  f: TSearchRec;
  err: integer;
begin
  FFullPath := APath;
  FDirName := AName;
  FLevel := ALevel;
  FType := AType;

  FIsOpen := false;

  case AType of
    FONFloppy: FHasSubDir := true;
    FONSpecial:FHasSubDir := false;
    else
      begin
	FHasSubDir := false;
	err := findfirst(APath + '\*.*',faDirectory, f);
	while err = 0 do
	  if ((f.Attr and faDirectory) = faDirectory) and (f.Name[1] <> '.')
	  then
	    begin
	      err := 1;
	      FHasSubDir := true;
	    end
	  else
	    err := findnext(f);
      end;
  end; { case..of }

  findclose(f);
end;

procedure TByFileNode.setOpen(value:boolean);
begin
  FIsOpen := value;
end;

procedure TByFileNode.SetFilterIndex(value: integer);
begin
  FFilterIndex := value;
end;

function TByFileNode.refresh: integer;
var
  f: TSearchRec;
  err: integer;
  AHasSub: Boolean;
  path : string;
begin
  Result := FENoError;
  if (FType and FONFolder) = FONFolder 
    then path := FFullPath
    else path := FFullPath+'\*.*';

  err := findfirst(path,faDirectory, f);
  findclose(f);
  if err <> 0
    then
      if err = -18
	then Result := FEDriveNotReady
	else Result := FEDirNotExist
    else
      if FType <> FONSpecial then
	begin
	  AHasSub := false;
	  err := findfirst(FFullPath + '\*.*',faDirectory, f);
	  while err = 0 do
	    if ((f.Attr and faDirectory) = faDirectory) and (f.Name[1] <> '.')
	    then
	      begin
		err := 1;
		AHasSub := true;
	      end
	    else
	      err := findnext(f);  { end while..do }

	  if FHasSubDir then
	    if not AHasSub then
	      begin
		FHasSubDir := false;
		Result := FENoSubDirExist;
	      end else else
	    if AHasSub then
	      begin
		FHasSubDir := true;
		Result := FESubDirExist;
	      end
      end;
end;

{$ifdef Win32}
procedure TByFileDirectory.CreateWnd;
var
  DriveIndex,
  DriveType   : Integer;
  DriveVolume : string[12];
  DrivePath   : string[2];

  PDrivePath  : Array[0..3] of Char;
  PDriveVolume: Array[0..8] of Char;
  MaxCompLength : DWORD;
  FileSysFlags : DWORD;
begin
  inherited CreateWnd;
  for DriveIndex := 0 to 31 do
  begin
    StrPCopy(PDrivePath, Char(65 + DriveIndex) + ':\');
    DriveType := GetDriveType(PDrivePath);
    if DriveType <> 1 then
    begin
      DrivePath := Char(65 + DriveIndex) + ':';
      DriveVolume:='';

      if DriveType = Drive_REMOVABLE then
        begin
          DriveType := FONFloppy;
          DriveVolume:= 'Floppy';
        end else
        begin
	  case DriveType of
	    Drive_FIXED : DriveType := FONFixed;
	    Drive_REMOTE: DriveType := FONNetDrv;
	    DRIVE_CDROM : DriveType := FONCDROM;
	    0: DriveType := FONUnknown;
	  end; {case..of}

	  FileSysFlags := FS_CASE_IS_PRESERVED;
	  GetVolumeInformation(PDrivePath,@PDriveVolume,255,nil,MaxCompLength,FileSysFlags,nil,0);

	  DriveVolume:= LowerCase(StrPas(PDriveVolume));
	  DriveVolume[1] := Char(Ord(DriveVolume[1]) - 32);
	end; {if}

      Items.AddObject('', TByFileNode.Create(DrivePath, DriveVolume + ' (' + DrivePath + ')', 0, DriveType));
    end;
  end;
end;

{$else}
procedure TByFileDirectory.CreateWnd;
var
  i,t: Integer;
  str : string[2];
  f : TSearchRec;
  DosErr: integer;
  V : string[12];
begin
  inherited CreateWnd;
  for i := 0 to 31 do
  begin
    t := GetDriveType(i);

    if t > 0 then
    begin
      str := Char(65 + i) + ':';
      V:='';

      if t = drive_Removable
        then
          begin
            t := FONFloppy;
            V:= 'Floppy';
          end
        else
          begin
            case t of
	      drive_Fixed:  t := FONFixed;
	      drive_Remote: t := FONNetDrv;
	      else t := FONUnknown;
	   end;

	   if findfirst(str + '\*.*',FaVolumeID, f) = 0 then
	   begin
	     V:= LowerCase(F.Name);
	     V[1] := Char(Ord(V[1]) - 32);
	   end;

	   FindClose(F);
	 end;

      Items.AddObject('', TByFileNode.Create(Str, V + ' (' + Str + ')', 0, t));
    end;
  end;
end;
{$endif Win32}

constructor TByFileDirectory.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  FOpen := TBitmap.Create;   fOpen.Handle := LoadBitmap(HInstance,'FD_FOLDER_OPEN');
  FClose := TBitmap.Create;  fClose.handle := LoadBitmap(HInstance,'FD_FOLDER_CLOSED');
  FFloppy := TBitmap.Create; ffloppy.Handle := LoadBitmap(HInstance,'FD_FLOPPY');
  FFixed := TBitmap.Create;  ffixed.Handle := LoadBitmap(HInstance,'FD_FIXED');
  FCDRom := TBitmap.Create;  fcdrom.Handle := LoadBitmap(HInstance,'FD_CDROM');
  FRemote := TBitmap.Create; fremote.Handle := LoadBitmap(HInstance,'FD_NET_DRV');

  IntegralHeight := false;
  Style := lbOwnerDrawFixed;
  OnDrawItem := DrawItem;

  FFileFolders := TByFileFolders.Create;
end;

destructor TByFileDirectory.Destroy;
begin
  FClose.Free;
  FOpen.Free;
  Ffloppy.Free;
  Ffixed.Free;
  Fcdrom.Free;
  Fremote.Free;

  FFileFolders.free;    
  inherited Destroy;
end;

procedure TByFileDirectory.SetFileFolders(value: TByFileFolders);
begin
  FFileFolders.Assign(value);
end;

procedure TByFileDirectory.DrawItem(Ctrl: TWinControl; Idx: Integer; R: TRect; State: TOwnerDrawState);
var
  x,w: integer;
  Rc: TRect;
  Bmp: TBitmap;

begin
  With Canvas do
  begin
    FillRect(R);

    with TByFileNode(Items.Objects[idx]) do
    begin
      if (FolderType and FONFOLDER) = FONFOLDER then
	begin
	  x:= (Level - 1)* 18 + 2;
	  if HasSubDir then
	    begin
	      Pen.Color := clBtnShadow;
	      rectangle(x+5, R.Top + 3, x+14, R.Top + 12);
	      Pen.Color := clBtnText;
	      moveto(x+7, R.Top + 7); lineTo(x+12, R.Top + 7);
	      if not IsOpen then
		begin
		  moveto(x+9, R.Top + 5);
		  lineTo(x+9,R.Top + 10);
		end;
	    end; { if HasSubDir }
	  x := x + 18;
	end else
	begin
	  x := 0;
	end; { if }

      case FolderType of
        FONFOLDER :
          if IsOpen
            then BrushCopy(Rect(x, R.Top+1, x+16, R.Top+15),FOpen,Rect(0,0,16,14),clBlue)
            else BrushCopy(Rect(x, R.Top+1, x+16, R.Top+15),FClose,Rect(0,0,16,14),clBlue);
	FONFloppy : BrushCopy(Rect(2, R.Top+1, 18, R.Top+15),FFloppy,Rect(0,0,16,14),clBlue);
	FONFixed,
        FONUnknown: BrushCopy(Rect(2, R.Top+1, 18, R.Top+15),FFixed,Rect(0,0,16,14),clBlue);
	FONCDROM  : BrushCopy(Rect(2, R.Top+1, 18, R.Top+15),FCDRom,Rect(0,0,16,14),clBlue);
	FONNetDrv : BrushCopy(Rect(2, R.Top+1, 18, R.Top+15),FRemote,Rect(0,0,16,14),clBlue);
        else
          begin
             Bmp := FFileFolders.Items[FilterIndex].icon;
             Rc := Rect(0,0,Bmp.Width-1, Bmp.Height-1);
             BrushCopy(Rect(x, R.Top+1, x+16, R.Top+15), Bmp ,Rc, clBlue);
          end;
      end;
      x := x + 20;

      TextOut(x, R.Top, DirName);
    end;
  end;
end;

function TByFileDirectory.Collapse: Boolean;
var
  idx : integer;
  CurrentNode  : TByFileNode;
  CurrentLevel : integer;
  temp: boolean;
  Err : integer;
begin
  idx := ItemIndex;
  CurrentNode := TByFileNode(Items.Objects[idx]);
  CurrentLevel := CurrentNode.Level;

  result := false;

  Err := CurrentNode.Refresh;
  case Err of
    FEDirNotExist:
	begin
	  { Current directory does not exist. Do recursive ByFileNode deletion }
	  if idx <> 0 then
	  repeat
	    dec(idx);
	  until (idx = 0) or (TByFileNode(Items.Objects[idx]).Level < CurrentLevel);
	  ItemIndex := Idx;
	  temp := Collapse;
	end;
    FEDriveNotReady:
	begin
	  { Drive not ready. Do recursive ByFileNode deletion }
	  if (idx <> 0) and (CurrentLevel <> 0) then
	  repeat
	    dec(idx);
	  until TByFileNode(Items.Objects[idx]).Level = 0;

	  try
	    while (idx < Items.Count -1) and (TByFileNode(Items.Objects[idx+1]).Level > 0) do
	    begin
	      Items.Objects[idx+1].free;
	      Items.Delete(idx+1);
	    end;
	  except
	    raise;
	  end;
	end;
    else
	if CurrentNode.IsOpen then
	try
	  while (idx < Items.Count -1) and (TByFileNode(Items.Objects[idx+1]).Level > CurrentLevel) do
	  begin
	    Items.Objects[idx+1].free;
	    Items.Delete(idx+1);
	  end;
	  Result := true;
	  CurrentNode.IsOpen := false;
	  SetScrollBar;
          Invalidate;
	except
          raise;
	end
  end { case .. of }
end;

function TByFileDirectory.Expand: Boolean;
var
  idx, NewPos : integer;
  F: TSearchRec;
  DosErr: integer;
  ParentNode  : TByFileNode;
  ParentPath  : string;
  ParentLevel : integer;
  NewNode, TempNode   : TByFileNode;
  FolderName: string;

  procedure ExpandFolders;
  begin
    ParentLevel := ParentNode.level;
    ParentPath :=  ParentNode.FullPath;
    DosErr := findfirst(ParentPath + '\*.*',faDirectory, f);

    while DosErr = 0 do
    begin
      if ((f.Attr and faDirectory) = faDirectory) and (f.name[1] <> '.') then
	begin
	  FolderName:= LowerCase(F.Name);
	  FolderName[1] := Char(Ord(FolderName[1]) - 32);

	  NewNode := TByFileNode.Create(ParentPath + '\' + FolderName{f.name},
		FolderName, ParentLevel + 1,FONFolder);

	  { sort. Could have used binary search to locate new position,
	    but I am too lazy }
	  NewPos := idx+1;
	  repeat
	    if NewPos < items.count then
	    begin
	      TempNode := TByFileNode(Items.Objects[NewPos]);
	      if (TempNode.FullPath < NewNode.FullPath) and
		 (TempNode.Level <> ParentLevel) {! See NOTE !}
		then inc(NewPos)
		else break
	    end else
	      break;
	  until false;

	  if NewPos < Items.Count
	    then Items.InsertObject(NewPos,'', NewNode)
	    else Items.AddObject('',NewNode);
	end;
      DosErr:=findnext(f);
    end; { while..do }

    findclose(f);
    Result := true;
  end;

  procedure ExpandSpecialFolders;
  var
     i: integer;
  begin
    {
       ParentLevel := ParentNode.level;
       ParentPath :=  ParentNode.FullPath;
    }

    for i := FFileFolders.Count - 1 downto 0 do
    begin
      DosErr := findfirst(ParentPath + '\' + FFileFolders.Items[i].filter, faArchive, f);
      while DosErr = 0 do
      begin
	if (f.Attr and faArchive) = faArchive then
	begin
	  FolderName:= LowerCase(F.Name);
	  FolderName[1] := Char(Ord(FolderName[1]) - 32);

	  NewNode := TByFileNode.Create(ParentPath + '\' + FolderName{f.Name},
		FolderName, ParentLevel + 1, FONSpecial);
          NewNode.FilterIndex := i;

	  NewPos := idx+1;
	  repeat
	    if NewPos < items.count then
	    begin
	      TempNode := TByFileNode(Items.Objects[NewPos]);
	      if (TempNode.FullPath < NewNode.FullPath) and (TempNode.FolderType = FOnSpecial)
		then inc(NewPos)
		else break
	    end else
	      break;
	  until false;

	  if NewPos < Items.Count
	    then Items.InsertObject(NewPos,'', NewNode)
	    else Items.AddObject('',NewNode);
	end;
        DosErr:=findnext(f);
      end; { while..do }
      findclose(f);
    end; { for..do }

  end;

begin
  idx := ItemIndex;
  ParentNode := TByFileNode(Items.Objects[idx]);
  result := false;

  DosErr := ParentNode.Refresh;

  { check if current folder still exist }
  if (DosErr = FEDirNotExist) or (DosErr = FEDriveNotReady)
    then Collapse
    else
      if not ParentNode.IsOpen then
	begin
	  if ParentNode.HasSubDir then ExpandFolders;
	  ExpandSpecialFolders;

	  if Result = true then
          begin
	    ParentNode.IsOpen := true;
	    ItemIndex := idx;

	    SetScrollBar;
	    Invalidate;
	  end;
      end;

  { NOTE:
     "if (TempNode.FullPath < NewNode.FullPath) then" would not work for this:

     d:                     d:
      |-- borland            |-- borland
      | |-+ delphi     ==>   |    |-- delphi    -- opened
      | |-+ delphi2          |    |-+ delphi2
      |-+ more folder        |      |-- bin     -- but sub-folders are put here
      |-+ more folder        |      |-- lib
  }
end;

procedure TByFileDirectory.SetScrollBar;
var
  Count,
  Index,
  L,
  MaxWidth  : integer;
begin
  MaxWidth := 0;
  Count := Items.Count - 1;

  for Index := 0 to Count do
    with TByFileNode(Items.Objects[index]) do
    begin
      if Level > 0 then
        L:= Level * 18 else
        L:= 0;
      L := Canvas.TextWidth(DirName) + L;
      if L > MaxWidth then MaxWidth := L;
    end;

  SendMessage(Handle, LB_SETHORIZONTALEXTENT, MaxWidth + 20, 0);
end;

procedure TByFileDirectory.DblClick;
begin
  with TByFileNode(Items.Objects[ItemIndex]) do
    if HasSubDir then
      if IsOpen
        then Collapse
        else Expand;
  inherited DblClick;
end;

procedure TByFileDirectory.Click;
begin
  if Assigned(FDirChange) then FDirChange(self);
  inherited Click;
end;

procedure TByFileDirectory.KeyDown(var Key: Word; Shift: TShiftState);
var
  Idx: integer;
  Count: integer;
begin
  idx := ItemIndex;
  if (Idx <> -1) and not ((ssCtrl in Shift) or (ssAlt in Shift)) then
  case Key of
    VK_Up:
	if idx > 0 then
	begin
	  ItemIndex := idx - 1;
	  if Assigned(FDirChange) then FDirChange(self);
	  Key := 0;
	end;
    VK_Down:
	if idx < Items.Count - 1 then
	begin
	  ItemIndex := idx + 1;
	  if Assigned(FDirChange) then FDirChange(self);
	  Key := 0;
	end;
    VK_Right, VK_Return:
	if Expand then
	begin
	  ItemIndex := Idx;
	  if Assigned(FDirChange) then FDirChange(self);
	  Key := 0;
	end;
    VK_Left :
	begin
	  Count := Items.Count;
	  if Collapse then
	  begin
	    ItemIndex := Idx;
	    if Assigned(FDirChange) then FDirChange(self);
	  end else
	  if Count = Items.Count then
	  with TByFileNode(Items.Objects[idx]) do
	    if Level > 0 then
	    begin
	      dec(Idx);
	      while (Items.Objects[Idx] as TByFileNode).Level >= Level do
		dec(Idx);
	      ItemIndex := Idx;
	      if Assigned(FDirChange) then FDirChange(self);
	    end;
	  Key := 0;
	end;
  end; { case..of }

  inherited KeyDown(Key, Shift);
end;

procedure TByFileDirectory.CNDrawItem(var Msg: TWMDrawItem);
begin
  with Msg.DrawItemStruct^ do
     if ItemState and ODS_FOCUS <> 0 then ItemState := ItemState and not ODS_FOCUS;
  inherited;
end;

procedure TByFileDirectory.SetDirectory(value: string);
var
  Idx, StrL : integer;
  CurrentDir: string; { current directory }
  TargetDir : string; { target directory }
  StrPos: integer;    { target[StrPos] = '\' }

  procedure GoToTargetDirectory;
  var
    ByFileNode: TByFileNode;
    TempInt: integer;
    NextDir: string;
  begin
    repeat
      if not TByFileNode(Items.Objects[Idx]).IsOpen then
        Expand;

      TempInt := StrPos;
      repeat
	inc(TempInt);
	if (TargetDir[TempInt] = '\') then
	  begin
	    inc(StrPos);
	    NextDir := Copy(TargetDir, StrPos, TempInt - StrPos);
	    break;
	  end else
        if TempInt = StrL then
          begin
	    NextDir := Copy(TargetDir, StrPos + 1, TempInt - StrPos);
	    break;
          end
      until false;

      StrPos := TempInt;

      repeat
	inc(idx);
      until UpperCase(TByFileNode(Items.Objects[Idx]).DirName) = NextDir;

      SendMessage(Handle, LB_SETCURSEL, Idx, 0);
    until StrPos >= StrL;
  end;

  { format value like "..\imagevue" or ".\imagevue" }
  procedure FormatPath(var path: string);
  var
    StrIdx,          { String Index }
    StrLen: integer; { String Length }
    StrOld,
    StrNew,
    StrCmp: string;  { String Component }
  begin
    StrLen := Length(path);

    { convert "c:/windows/system" to "c:\windows\system" }
    repeat
      StrIdx := Pos(path,'/');
      if StrIdx = 0 then
	break else
	Path[StrIdx] := '\';
    until false;

    StrOld := path;

    repeat
      StrIdx := Pos(StrOld,'\');
      if StrIdx <> 0 then
	begin
	  StrCmp := Copy(StrOld,1, StrIdx)
	end else
	begin
	end;
    until false;

    path := StrNew;
  end;

begin
  CurrentDir := GetDirectory;

  StrL := length(value);
  if value[StrL] = '\' then
    begin
      dec(StrL);
      TargetDir := copy(value,2,StrL);
    end else
      TargetDir := value;

  if TargetDir[2] <> ':' then
    begin
      if TargetDir[1] <> '\' then
	CurrentDir := CurrentDir + '\';

      TargetDir := UPPERCASE(TargetDir);
      TargetDir := CurrentDir + TargetDir;

      StrL := Length(TargetDir);
      Idx := ItemIndex;
      StrPos := Length(CurrentDir);

    end else
    begin
      TargetDir := UPPERCASE(TargetDir);
      if (TargetDir[1] >= 'A') and (TargetDir[1] <= 'Z') then
      begin
	Idx := 0;
	repeat
	  with TByFileNode(Items.Objects[idx]) do
	    if (Level = 0) and (FullPath[1] = TargetDir[1]) then
	      break else
	      inc(Idx);
	until false or (Idx = Items.Count);
	StrPos := 3;
      end;
    end;

{  FormatPath(TargetDir); <-- crashes system for some reason}

  if CurrentDir <> TargetDir then
  begin
    if FileExists(TargetDir) then
      begin
	Items.EndUpdate;
	SendMessage(Handle, LB_SETCURSEL, Idx, 0);
	GotoTargetDirectory;

	if Assigned(FDirChange) then
	  FDirChange(self);
	Items.BeginUpdate;
      end else
	raise Exception.Create('Invalid directory: ' + TargetDir)
  end;
end;

function TByFileDirectory.GetDirectory: string;
begin
  if ItemIndex <> -1
    then Result := TByFileNode(Items.Objects[ItemIndex]).FullPath
    else Result := '';
end;

procedure TByFileDirectory.DefineProperties(Filer: TFiler);
begin
  inherited DefineProperties(Filer);

  Filer.DefineBinaryProperty('FileFoldersData', ReadData, WriteData,
        true);
end;

procedure TByFileDirectory.ReadData(Stream: TStream);
begin
  FFileFolders.ReadData(Stream);
end;

procedure TByFileDirectory.WriteData(Stream: TStream);
begin
  FFileFolders.WriteData(Stream);
end;

end.
